; Determine the offset in magnitude (third column) between two catalogs (positions first two columns)

; This compares the STAR CATALOG magnitudes of the two tables.  The zeropoints in the header(s) are NOT RELEVANT.


; format:  getzptoffset, comparisonfiles, [referencefile], [rms=rms], nmatch=nmatch, getref=getref

; omit referencefile to use SDSS/Landolt.

; zpt will add the returned value to the image zeropoint (i.e., return the zeropoint, assuming that 
;   the 'reference' file is correctly calibrated.)

function getzptoffset, comparisonfiles, referencefile, rms=rms, unc=unc, nmatch=nmatch, zptabs=zptabs, getref=getref, filter=filter, nir=nir

   if n_elements(referencefile) gt 0 then reffile = referencefile else reffile = ''
   if n_elements(comparisonfiles) eq 0 then begin
      print, 'No file given!'
      return, -1
   endif
   compfiles = comparisonfiles

   if reffile ne '' then begin  
      if strpos(reffile,'.fits') gt -1 then reffile = repstr(reffile,'.fits','.cat')
      if file_test(reffile) eq 0 then begin
         print, 'Reference file ', reffile, ' does not exist.'
         return, !values.f_nan
      endif
   endif

   for f = 0, n_elements(compfiles)-1 do begin
      if strpos(compfiles[f],'.fits') gt -1 then compfiles[f] = repstr(compfiles[f],'.fits','.cat')
      if file_test(compfiles[f]) eq 0 then begin
         print, 'Comparison file ', compfiles[f], ' does not exist.'
         return, !values.f_nan
      endif
   endfor

   if reffile ne '' then begin
     if strpos(reffile,'.sdss') gt 0 then begin  ; This is a somewhat dissatisfying solution.  Need to find a way to look into the file
        calfile = reffile                      ; to determine the format (printcatalog SDSS/2MASS table versus fieldphot table)
        catalog = 'sdss'
     endif else begin
        ref = readcat(reffile, h)  ; if it's a printcatalog-type multifilter reference file this will FAIL
        catalog = reffile
     endelse
   endif else begin
      ; no reffile specified - look it up online via the footprint of the image

      compimages = repstr(compfiles,'.cat','.fits')  
      compimages = repstr(compimages,'.cal','.fits') 
      compimages = repstr(compimages,'.caz','.fits') 
      if strpos(compimages,'.fit') le 0 then begin
         print, 'Need to access FITS image to do online reference lookup.'
      endif
      if keyword_set(nir) eq 0 then begin
        catalog = 'landolt'
        printcatalog, compimages, catalog, outfile='temp.cat', /quiet, /unc, count=count
        if count eq 0 then begin
          catalog = 'sdss'
          printcatalog, compimages, 'sdss', outfile='temp.cat', /quiet, /unc, count=count
        endif
      endif else begin
        catalog = '2mass'
        printcatalog, compimages, catalog, outfile='temp.cat', /quiet, /unc, count=count
      endelse
      calfile = 'temp.cat'

      if count eq 0 then begin
         catalog = ''
         print, 'No calibrations of this field.'
         return, !values.f_nan
      endif

      ;comp = readcat(compfiles[f], hc)
      ;filter = sxpar(comp,'FILTER')
      ;cat = getlandolt(avgra, avgdec, 400., count=n)
      ;if n lt 1 then $
      ;cat = getsdss(avgra, avgdec, 400., count=n)
      ;if n lt 1 then print, 'No reliable photometric catalog is available for this field.'
      ;if n lt 1 then return
      ;if catalog eq '2mass' then cat = get2mass(avgra, avgdec, 400.)      
      
   endelse

   zpt = fltarr(n_elements(compfiles))
   dzpt = fltarr(n_elements(compfiles))
   dzptrms = fltarr(n_elements(compfiles))
   dzptunc = fltarr(n_elements(compfiles))

   for f = 0, n_elements(compfiles)-1 do begin
      comp = readcat(compfiles[f], hc)
      zpt[f] = sxpar(hc, 'ZEROPT')  ; note that this is (now) not used until the end and only if zptabs is set

      if reffile eq '' then begin
         if n_elements(filter) eq 0 then begin
            filter = sxpar(hc,'FILTER')               ; need more robust treatment of this...
            if filter eq 'RG850' then filter = 'z'  
            if filter eq 'G' then filter = 'g'  
            if filter eq 'U' then filter = 'u'  
            if filter eq 'Clear' then filter = 'R'
            if filter eq 'clear' then filter = 'R'
            if filter eq 'upr' then filter = 'u'
            if filter eq 'gpr' then filter = 'g'
            if filter eq 'rpr' then filter = 'r'
            if filter eq 'ipr' then filter = 'i'
            if filter eq 'zpr' then filter = 'z'
         endif
         ; The readcalibration call is down in this block because it's filter specific, and an array of comparison files of different filters may be passed in.
         ref = readcalibration(calfile, magcol=magcolnum, unccolnum=unccolnum, filter=filter)
         w = where(ref.unc lt 0.28, n)
         if n gt 3 then ref = ref[w]
      endif

      if n_elements(h) gt 1 then begin
      if sxpar(h,'ZEROPT') ne sxpar(hc,'ZEROPT') then begin
         ;;print, 'Different zeropoints employed, adjusting.'
         ;comp.mag += (sxpar(h,'ZEROPT') - sxpar(hc,'ZEROPT'))  ; this might be the wrong sign
         ;I think we actually do not want to do this.
      endif
      if sxpar(h,'RZPT') ne sxpar(hc,'RZPT') then begin
         if abs(sxpar(h,'RZPT')-sxpar(hc,'RZPT')) / ((sxpar(h,'RZPT')+sxpar(hc,'RZPT'))/2.) gt 0.01 then $   ; require at least a 1% diff
         print, 'WARNING - Zeropoints were calculated with different apertures ('+fpr(sxpar(h,'RZPT'),2.3)+' vs. '+fpr(sxpar(hc,'RZPT'),2.3)+')'
      endif
      if clip(sxpar(h,'FILTER')) ne clip(sxpar(hc,'FILTER')) then begin
         print, 'WARNING - Images were taken with different filters!'
      endif
      endif

      matchcat, ref.ra, ref.dec, comp.ra, comp.dec, rmatch, cmatch, mindist=1./3600, nmatch=nmatch

      if nmatch eq 0 then begin
         print, 'No matches!!'
         dzpt[f] = !values.f_nan
         dzptunc[f] = !values.f_nan
      endif else begin

         ;for i = 0, n_Elements(rmatch)-1 do begin
         ;   print, ref[rmatch[i]].mag, comp[cmatch[i]].mag, ref[rmatch[i]].mag-comp[cmatch[i]].mag
         ;endfor

         dmagarr = ref[rmatch].mag - comp[cmatch].mag
         dmagarrunc = sqrt(ref[rmatch].unc^2 + (comp[cmatch].unc)^2)
         sigclipstats, dmagarr, median=dmagmed, stdev=dmagstdev  ; should really fold errors into this
         dzpt[f] = wmean(dmagarr, dmagarrunc, sigclip=3., meanunc=meanunc, nrej=nrej)
         ;print, '          ', dmagarr
         ;print, '       +/-', dmagarrunc
         ;print, '        > ', dmagmed, dzpt[f]
         ;nmatch -= nrej
         if n_elements(dmagarr) gt 1 then dzptrms[f] = stdev(dmagarr-dzpt[f])^2 else dzptrms[f] = 0.
         dzptunc[f] = meanunc ; would be useful to remove the photometric error component from this scatter
      endelse


   endfor

   ; only return arrays if arrays given 
   if n_elements(zpt) eq 1 then zpt = zpt[0]
   if n_elements(dzpt) eq 1 then dzpt = dzpt[0]
   if n_elements(dzptrms) eq 1 then dzptrms = dzptrms[0]
   if n_elements(dzptunc) eq 1 then dzptunc = dzptunc[0]

   rms = dzptrms
   unc = dzptunc
   getref = catalog
   if keyword_set(zptabs) eq 0 then return, dzpt
   if keyword_set(zptabs)      then return, zpt+dzpt
end

