;+
; NAME: lrisprepare.pro
;
; PURPOSE: 
;       All-purpose program to perform basic operations on an LRIS
;       image (or 2D spectrum) including bias subtraction, flagging
;       of bad pixels, header standardization, provisional astrometry,
;       etc.
;       Auto-detects whether it is red or blue, which LRIS version
;       it is, and whether it is imaging or spectroscopy.
;       This the latest, final, catch-all version.
;       It does not yet work with old-format images, but should work
;       with all modern LRIS imaging/spectroscopy regardless of amp
;       status, binning, etc.
;
; CALLING SEQUENCE: 
;       lrisprepare, filename
;
; INPUTS:
;	filename = name of the multi-HDU FITS file to prepare,
;                  or array of filenames, or file w/list of filenames
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;       split       - write out left and right chips separately (default)
;       merge       - keep left and right chips together
;       multi       - like split but make a multi-extension file
;       rightonly 
;       leftonly
;       crops       - crop instruction
;       xcrop       - crop locations
;       lcrop       -
;       rcrop       -
;       ycrop       -
;       outname     - specify output file to write to disk
;       loutname
;       routname
;       prefix      - specify output prefix
;       lprefix     -
;       rprefix     -
;       suffix      - specify output suffix
;       lsuffix     -
;       rsuffix     -
;       nobias
;       scalarbias
;       gain        - gain correction for all four (two if LRISR1) amps
;       flag        - do not use?? implemented inconsistently
;       fixdefect   - replaces flagged values with interpolations
;       verbose
;       namefix     - correct names using a catalog
;       timer       - print out processing times
;
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;       header (?)
; 
; OUTPUTS:
;       No formal outputs.
;       The prepared images are written to disk.
;
; REQUIREMENTS:
;       - Requires the IDL Astronomy User's Library routines 
;       (http://idlastro.gsfc.nasa.gov/) 
;       - Requires readmhdu.pro to process LRIS images from 2009/later.

; Note - now attempts to add physical coordinates based on binned data.
; However, there is a genuine physical shift for binned observations - when binned 2x2,
; an extra row of dead pixels is read out along the edge of the gap (on the r chip and 
; looks like on the l chip too.)  This applies to R3; I don't know if it applies to R2.

;----------------------------------------------------------------------------------------
pro lrisprepare, filename, split=split, merge=merge, multi=multi, rightonly=rightonly, leftonly=leftonly, $
        crops=crops, xcrop=xcrop, lcrop=lcrop, rcrop=rcrop, ycrop=ycrop, notranspose=notranspose, $
        outname=outname, loutname=loutname, routname=routname, $
        prefix=prefix, lprefix=lprefix, rprefix=rprefix, $
        suffix=suffix, lsuffix=lsuffix, rsuffix=rsuffix, $
        nobias=nobias, scalarbias=scalarbias, gain=gain, verbose=verbose, $
        fixdefect=fixdefect, flag=flag, namefixfiles=namefixfiles, header=header, timer=timer, seq=seq, $
        requirecompleteheader=requirecompleteheader, newobj=newobj
   ; arguments repeated once

common lrisconfig, lrisautopath, refdatapath, defaultspath
common lrisbpmblock, bpm1, bpm2, bpm3 
common lrisprepareprev, mask, prevcamera, prevversion, prevpane, prevxbin, prevybin
common lrisbadarc, badarcdir, badarclist
common lrisdrop, dropcolumns

if n_elements(refdatapath) eq 0 then bpmpath = '/home/dperley/progs/idl/lris/refdata/' else bpmpath = refdatapath
if strlen(bpmpath) gt 0 then if strmid(bpmpath,strlen(bpmpath)-1,1) ne '/' then bpmpath += '/'

if keyword_set(merge) and keyword_set(split) then begin
   print, 'Cannot both split and merge.'
   return
endif
if keyword_set(merge) eq 0 and keyword_set(split) eq 0 then begin
   split = 1   
endif
merge = keyword_set(merge)
split = keyword_set(split)
multi = keyword_set(multi)

if n_elements(crops) eq 0 and $
   n_elements(xcrop) eq 0 and n_elements(lcrop) eq 0 and n_elements(rcrop) eq 0 and n_elements(ycrop) eq 0 then crops = 'auto'

; ---------- Process input filename(s), call recursively if necessary----------

; Check for empty filename
if n_elements(filename) eq 0 then begin
   print, 'No filename specified.'
   return
endif

; Check for array input
if n_elements(filename) gt 1 then begin
   files = filename
endif

; Check for list input
if n_elements(filename) eq 1 then begin
   filearr = strsplit(filename,'.', /extract)
   fileroot = filearr[0]
   fileext = filearr[1]
   if fileext eq 'cat' or fileext eq 'lis' or fileext eq 'list' or fileext eq 'txt' then begin
     ;files = grabcolumn(filename,0,/str)
     readcol, filename, files, format='a'
   endif
endif


; Check for wildcard input
if n_elements(filename) eq 1 then begin
   starpos = strpos(filename,'*')
   qpos = strpos(filename,'?')
   if starpos ge 0 or qpos ge 0 then begin
      files = file_search(filename)
      if n_elements(files) eq 1 then begin
        if files[0] eq '' then print, 'Cannot find any files matching ', filename
        return
      endif
   endif
endif

; Check for array filename, loop recursively if necessary
if n_elements(files) gt 1 then begin
  for f = 0, n_elements(files)-1 do begin
    lrisprepare, files[f], split=split, merge=merge, rightonly=rightonly, leftonly=leftonly, $
        crops=crops, xcrop=xcrop, lcrop=lcrop, rcrop=rcrop, ycrop=ycrop, notranspose=notranspose, $
        outname=outname, loutname=loutname, routname=routname, $
        prefix=prefix, lprefix=lprefix, rprefix=rprefix, $
        suffix=suffix, lsuffix=lsuffix, rsuffix=rsuffix, $
        nobias=nobias, scalarbias=scalarbias, gain=gain, verbose=verbose, $
        fixdefect=fixdefect, flag=flag, namefixfiles=namefixfiles, timer=timer  ; seq won't work for an array
  endfor
  return
endif


print, filename, format='($,A)'


; ---------- Read data and process header information ----------

; Read in the data and header

starttime = systime(/seconds) 
linebias = keyword_set(scalarbias) eq 0

check = file_test(filename)
if check eq 0 then begin
  print, ' not found!'
  return
endif

h = headfits(filename)

mjd = float(sxpar(h,'MJD-OBS'))
if mjd gt 0 and mjd lt 49000 then begin
   print
   print, ' WARNING: Observation date does not make sense!'
   print, '   MJD-OBS  '+clip(sxpar(header,'MJD-OBS'))
   print, '   DATE-OBS '+clip(sxpar(header,'DATE-OBS'))
   print, ' Assuming this is recent data.  If observations predate 2011 this will cause problems.'
   lrisversion = 3
endif
if mjd gt 49000 and mjd lt 54952 then lrisversion = 1
if mjd ge 54952 and mjd lt 55562 then lrisversion = 2
if mjd ge 55562 then lrisversion = 3
if n_elements(lrisversion) eq 0 then begin
   print, '[No MJD entry in header: assuming LRIS3]', format='($,A)'
   lrisversion = 3
endif

if sxpar(h, 'NAXIS') eq 2 then begin
   ;fits_info, filename, /SILENT, N_ext=n_ext, extname=extname  (an alternative way to do this)

   if sxpar(h,'INSTRUME') eq 'LRIS' and lrisversion eq 1 then begin
      ; Old LRIS red camera
      array = readlrisr1(filename, header=header, nobias=nobias, verbose=verbose) 
   endif else begin
      ; Old-style simple LRIS file format
      array = readoldlris(filename, header=header, nobias=nobias, linebias=linebias, verbose=verbose) 
   endelse
   ; the two lrisread functions above are nearly identical and could be unified.

endif else begin

   ; Modern multi-extension LRIS file format
   array = readmhdufitsdap(filename, header=header, nobias=nobias, linebias=linebias, verbose=verbose) ; do gain later

endelse
delvarx, h ; not really necessary, but useful for debugging

sin = size(array) ; input size
nxin = sin[1]
nyin = sin[2]

; for old-style LRIS, at this point we will have to indicate where the amps are manually.

if nxin eq 4096 and nyin eq 512 and sxpar(header,'INSTRUME') eq 'LRISBLUE' and clip(sxpar(header,'SLITNAME')) eq 'direct' then begin
   print, ' - MIRA file; skipping.'
   return
endif


readtime = systime(/seconds)  ; mark time after file read from disk with readmhdu



; Get essential info about the data structure: version, camera, binning, windowing, lmode



instrume = clip(sxpar(header,'INSTRUME'))
if instrume eq 'LRIS'     then camera = 'r'
if instrume eq 'LRISBLUE' then camera = 'b'
if camera eq 'r' then sxaddpar, header, 'CAMERA', 'red'
if camera eq 'b' then sxaddpar, header, 'CAMERA', 'blue'
sxaddpar, header, 'LRISVER', lrisversion
if camera eq '' then begin
   print, ': unrecognized camera (INSTRUME = ', instrume, ')'
   return
endif

binning = clip(sxpar(header,'BINNING'))
newbinning = repstr(binning, ' ', '') ; remove spaces
sxaddpar, header, 'BINNING', newbinning
bin = fix(strsplit(binning,',',/extract))
xbin = bin[0]
ybin = bin[1]
sxaddpar, header, 'XBIN', xbin, after='BINNING'  ; actually Y, but we rotate it later for spectroscopy...
sxaddpar, header, 'YBIN', ybin, after='BINNING'

panestr = clip(sxpar(header,'PANE'))
if panestr eq '' or panestr eq '0' then begin
   ; PANE is always in unbinned coordinates:  x0, y0, dx, dy
   ; WINDOW is in binned coordinates:         0, x0, [y0], dx, dy
   windowstr = clip(sxpar(header,'WINDOW',count=ctwindow))
   windowvals = fix(strsplit(windowstr,',',/extract))
   if ctwindow eq 0 or n_elements(windowvals) lt 4 then begin
      print, 'ERROR:  No windowing keyword data?  Cannot determine amplifier boundaries.'
      return
   endif
   if lrisversion eq 1 and camera eq 'r' then $
   windowvals[2] = 2048/xbin - windowvals[4] - windowvals[2] ; seems this is defined from the right? at least it was in 2007
   if lrisversion eq 1 and camera eq 'b' then $
   windowvals[3] *= 2  ; was 1,0,0,2048,4096 even in unbinned data (not sure about binned data)
   panevals = intarr(4)
   panevals[0] = windowvals[1] * xbin
   panevals[1] = windowvals[2] * ybin
   panevals[2] = windowvals[3] * xbin
   panevals[3] = windowvals[4] * ybin
   panestr = ''
   for p = 0, 3 do panestr += clip(panevals[p]) + (p lt 3 ? ',' : '')
   sxaddpar, header, 'PANE', panestr
endif
pane = fix(strsplit(panestr,',',/extract))

lmode = '' ; mode based on disperser
if camera eq 'b' and clip(sxpar(header,'GRISNAME'))  eq 'clear'  then lmode = 'img'
if camera eq 'r' and clip(sxpar(header,'GRANAME'))  eq 'mirror' then lmode = 'img'
if camera eq 'b' and strpos(sxpar(header,'GRISNAME'),'/') gt 0   then lmode = 'spec'
if camera eq 'r' and strpos(sxpar(header,'GRANAME'),'/') gt 0    then lmode = 'spec'

smode = '' ; mode based on slit
slit =  strlowcase(clip(sxpar(header, 'SLITNAME')))
if slit eq 'direct' then smode = 'img'
if strpos(slit,'_') gt 0 then smode = 'spec'

if lmode eq 'img' and smode eq 'spec' then begin
   ; or smode eq 'img' and lmode eq 'spec'  -- this is now OK, slitless spectrum
   print
   print, '     WARNING: Mixed-up grating/slit settings!  Ignoring file...'
   return
endif
if lmode eq '' and smode eq '' then begin
   print
   print, '     WARNING: grating and slit keywords corrupt, absent, or unusual.'
  if camera eq 'b' then print, '      GRISNAME:', clip(sxpar(header,'GRISNAME')), '    SLITNAME:', clip(sxpar(header,'SLITNAME')), '  FILTER:', clip(sxpar(header,'BLUFILT')), '  :', lmode, '?'
  if camera eq 'r' then print, '       GRANAME:', clip(sxpar(header,'GRANAME')), '  SLITNAME:', clip(sxpar(header,'SLITNAME')), '  FILTER:', clip(sxpar(header,'REDFILT')), '  :', lmode, '?'
   print, '              Skipping...'
   return
endif
if lmode eq '' and smode ne '' then lmode = smode


  ;lmode = 'img'
  ;print
   ;endif

  ; Check for missing keywords that are required for the pipeline to work
  ; (lots of instrument keywords seem to not ever be missing so aren't included here)

   hdate = sxpar(header,'DATE', count=checkdate, /silent)
   ; htime = sxpar(header,'UT', count=checktime)
   hdateobs = sxpar(header,'DATE-OBS', count=checkdateobs, /silent)
   if checkdate gt 0 and checkdateobs eq 0 and lrisversion eq 1 then sxaddpar, header, 'DATE-OBS', clip(hdate)

   checkkeysboth =  ['OBJECT','TARGNAME','DOMEPOSN','INSTRUME',$,
                    'RA','ROTMODE','ROTPOSN','ELAPTIME',$
                    'AIRMASS','AZ','DEC','EL','HA','MJD-OBS','ST','DATE-OBS']
   checkkeysim = ['REDFILT','BLUFILT']
   checkkeyssp = ['WAVELEN','GRISNAME','GRANAME','SLITNAME']
   if lmode eq 'img' or smode eq 'img' then checkkeys = [checkkeysboth, checkkeysim]
   if lmode eq 'spec' or smode eq 'spec' then checkkeys = [checkkeysboth, checkkeyssp]
                ; also important but have changed convention: UT/UTC, LAMPS/(lamp names), WINDOW/PANE
                ; dome lamp keywords? 
   missing = ['']
   for k = 0, n_Elements(checkkeys)-1 do begin
      test = sxpar(header, checkkeys[k], count=ct, /silent)
      if ct eq 0 then missing = [missing, checkkeys[k]]
   endfor
   if n_elements(missing) gt 1 then begin
       misstr = ''
       for km = 0, n_elements(missing)-1 do misstr += ','+missing[km]
       misstr = strmid(misstr,2)
       print, ' - missing the following keywords: '+misstr, format='($,A)'
       if keyword_set(requirecompleteheader) then begin
          print, ' - not processing'
          return
       endif
   endif

;targname = sxpar(header, 'TARGNAME', count=ct)
;if ct eq 0 then begin
;   print, '[TARGNAME keyword missing from header!]', format='($,A)'
;   sxaddpar, header, 'TARGNAME', 'unknown'
;endif

; Tidy up the header by removing unhelpful information and 
; reorganizing some keywords.
; If anyone recognizes keywords here that do contain useful information
; for an observer, please let me know.

sxdelpar, header, ''           ; remove blank lines
sxdelpar, header, 'COMMENT' 
sxdelpar, header, 'SIMULATE'
sxdelpar, header, 'PMFM'
adcdelpars = 'ADC'+['NEED','PRENA','BYP','CAL','CMT','CON','DCC','ENC','ERR','GTE','LCK','LIM','MOD','RAW',$
                    'RCD','STA','SVA','SVR','SVX','TVA','TVX','VAL','VAX','VEL','VEX']
sxdelpar, header, adcdelpars
sxdelpar, header, 'CTRL0ERR'
sxdelpar, header, 'CTRL0MSG'
sxdelpar, header, 'DISP0ERR' 
sxdelpar, header, 'DISP0STA'
sxdelpar, header, 'POW24V'
sxdelpar, header, 'PRECOL'   ; deal with prepix, etc. which have been removed already
sxdelpar, header, 'PREROW'
sxdelpar, header, 'PRELINE'
sxdelpar, header, 'POSTLINE'
sxdelpar, header, 'ERASLINE'
sxdelpar, header, 'KEEPPREP'
sxdelpar, header, 'PREFLUSH'
sxdelpar, header, 'OVRFLUSH'
sxdelpar, header, 'DETCNFID'
sxdelpar, header, 'VIDINACT'
for p = 1, 16 do begin
  sxdelpar, header, 'PANE'+clip(p)
endfor
sxdelpar, header, 'TV1FILT'
sxdelpar, header, 'TV2FILT'
sxdelpar, header, 'TV1FPOS'
sxdelpar, header, 'UTBTEMP'
sxdelpar, header, 'UTBP30V'
sxdelpar, header, 'UTBP15V'
sxdelpar, header, 'UTBM15V'
sxdelpar, header, 'UTBP5V'
sxdelpar, header, 'AUTOERAS'
sxdelpar, header, 'AUTOSHUT'
sxdelpar, header, 'ERASECNT'
sxdelpar, header, 'MPPMODE'
sxdelpar, header, 'NSUBINT'
sxdelpar, header, 'ANTIBLM'
sxdelpar, header, 'ABFREQ'
sxdelpar, header, 'POSTPIX'
sxdelpar, header, 'TUBETEMP'
sxdelpar, header, 'DECCNFID'
sxdelpar, header, 'VIDINACT'
sxdelpar, header, 'CHECKSUM'
sxdelpar, header, 'DATASUM'
sxdelpar, header, 'SHBLOPTM'
sxdelpar, header, 'SHBLCLTM'
sxdelpar, header, 'SHLATOPN'
sxdelpar, header, 'SHLATCLO'
sxdelpar, header, 'SHOPNCNT'
sxdelpar, header, 'SHCLOCNT'
sxdelpar, header, 'TEMP1C'
sxdelpar, header, 'TEMP2C'
sxdelpar, header, 'TEMPELC'
sxdelpar, header, 'CCDSUM'

sxrenamepar, header, 'GRISM', 'GRISNUM'    ; these cause confusion.
sxrenamepar, header, 'DICHROIC', 'DICHNUM'
sxrenamepar, header, 'GRATING', 'GRANUM'
sxrenamepar, header, 'OBJECT', 'TEMPOBJ' ; remove duplicate keyword
sxrenamepar, header, 'TEMPOBJ', 'OBJECT'

sxmovepar, header, 'TEMPDET'
sxmovepar, header, 'TEMPSET'

sxmovepar, header, 'DEC', after='DATE-OBS'
sxmovepar, header, 'RA', before='DEC'
sxmovepar, header, 'EQUINOX', after='DEC'
sxmovepar, header, 'HA', before='AIRMASS'
sxmovepar, header, 'OBJECT', after='TARGNAME'

if clip(sxpar(header, 'UTC')) eq '0' then sxaddpar, header, 'UTC', sxpar(header, 'UT'), after='UT'
if clip(sxpar(header, 'UT')) eq '0' then sxaddpar, header, 'UT', sxpar(header, 'UTC'), after='UTC'

sxaddpar, header, 'BZERO', 0 
sxaddpar, header, 'BSCALE', 1
sxaddpar, header, 'NAXIS1', 0, after='NAXIS' ; just to save the location
sxaddpar, header, 'NAXIS2', 0, after='NAXIS1'


; Compatibility of lamps format (must be done before namefix, which uses lamp info)
lamps = sxpar(header, 'LAMPS', count=lampct)
if lampct eq 0 then begin  ; compatibility with older lamp format
      hg = clip(sxpar(header, 'MERCURY')) eq 'on'
      neon = clip(sxpar(header, 'NEON')) eq 'on'
      ar = clip(sxpar(header, 'ARGON')) eq 'on'
      cd = clip(sxpar(header, 'CADMIUM')) eq 'on'
      zn = clip(sxpar(header, 'ZINC')) eq 'on'
      hal = clip(sxpar(header, 'HALOGEN')) eq 'on'
      kr = clip(sxpar(header, 'KRYPTON')) eq 'on'
      xe = clip(sxpar(header, 'XENON')) eq 'on'
      fear = clip(sxpar(header, 'FEARGON')) eq 'on'
      de = clip(sxpar(header, 'DEUTERI')) eq 'on'
      lampstr = clip(fix(hg))+','+clip(fix(neon))+','+clip(fix(ar))+','+clip(fix(cd))+','+clip(fix(zn))+','+clip(fix(hal)) $
                        +','+clip(fix(kr))+','+clip(fix(xe))+','+clip(fix(fear))+','+clip(fix(de))
      sxaddpar, header, 'LAMPS', lampstr, before='MERCURY'
endif


result = 0
if n_elements(namefixfiles) gt 0 then begin
   namefix, header, namefixfiles, result=result
endif
if n_elements(newobj) gt 0 and result eq 0 then begin
   origobj = sxpar(header,'OBJECT')
     print, ' ['+clip(origobj)+' -> '+newobj+'] ', format='($,A)'
   sxaddpar, header, 'OBJECT', newobj
endif

if sxpar(header, 'EXPTIME') eq 0 then begin ; swarp needs this
    if sxpar(header, 'ELAPTIME') gt 0 then begin
       sxaddpar, header, 'EXPTIME', sxpar(header, 'ELAPTIME'), before='ELAPTIME'
    endif else begin
       if sxpar(header, 'TTIME') gt 0 then $
          sxaddpar, header, 'EXPTIME', sxpar(header, 'TTIME'), before='TTIME'
    endelse
endif 


; Saturation on the help page keeps changing.  Currently claims both are 60k.
if camera eq 'b' then sxaddpar, header, 'SATURATE', 60000, before='BINNING'
if camera eq 'r' and lrisversion eq 1 then sxaddpar, header, 'SATURATE', 60000, before='BINNING'
if camera eq 'r' and lrisversion ge 2 then sxaddpar, header, 'SATURATE', 60000, before='BINNING'

if camera eq 'b' then sxaddpar, header,'FILTER', sxpar(header,'BLUFILT'), before='BLUFILT'
if camera eq 'r' then sxaddpar, header,'FILTER', sxpar(header,'REDFILT'), before='BLUFILT'
if camera eq 'b' and clip(sxpar(header, 'GRISNAME'),2) eq 'NB' then begin
   ; narrowband blue filers are in grism carousel
  if clip(sxpar(header,'BLUFILT')) eq 'clear' then sxaddpar, header,'FILTER', sxpar(header,'GRISNAME')  $
                                              else sxaddpar, header,'FILTER', sxpar(header,'GRISNAME')+sxpar(header,'BLUFILT')
endif

if lmode eq 'img' then begin
   filtcwav = filtwv(sxpar(header,'FILTER'), fwhm=filtfwhm)
   if filtcwav gt 1 then sxaddpar, header, 'FILTCWAV', filtcwav, after='FILTER'
   if filtfwhm gt 1 then sxaddpar, header, 'FILTFWHM', filtfwhm, after='FILTCWAV'
endif


if camera eq 'b' then sxdelpar, header, ['REDFILT','REDFNUM','REDFOCUS']
if camera eq 'r' then sxdelpar, header, ['BLUFILT','BLUFNUM','BLUFTRAN','BTRAN','BLUFOCUS']


if camera eq 'r' then begin
   if lrisversion eq 1 then pxscale = 0.2115*(xbin) $
                       else pxscale = 0.135*(xbin)
endif
if camera eq 'b' then pxscale = 0.1357*(xbin)

sxaddpar, header, 'PXSCALE', pxscale, 'arcsec/pix', after='WAVELEN'
sxaddpar, header, 'OBSERVAT',"Keck", before='TELESCOP'
sxaddpar, header, 'SLITPA', (float(sxpar(header, 'ROTPOSN'))+90) mod 360., 'Slit or right chip CCW from North', before='ROTPOSN'

sxaddpar, header, 'RDNOISE', 4, before='BINNING' ; default as placeholder
if camera eq 'b' then  sxaddpar, header, 'RDNOISE', 3.6    ;actually varies between chips/amps

; --- Grating/grating angle settings -----
  ; currently not used.
if n_elements(seq) gt 0 then begin
  sxaddpar, header, 'DICHSEQF', seq.dichf, 'First file after most recent DICHNAME change'
  sxaddpar, header, 'RFLTSEQF', seq.rfiltf, 'First file after most recent REDFILT change' 
  sxaddpar, header, 'BFLTSEQF', seq.bfiltf, 'First file after most recent BLUFILT change' 
  sxaddpar, header, 'GRATSEQF', seq.gratf, 'First file after most recent GRANAME change' 
  sxaddpar, header, 'GRANSEQF', seq.granglef, 'First file after most recent GRANGLE change' 
endif



; --- Sun/moon ephemeris ---

mjd = sxpar(header, 'MJD-OBS')
jd = double(mjd) + 2400000.5
sxaddpar, header, 'JD', jd, after='MJD-OBS'
jdmid = jd + sxpar(header,'EXPTIME')/3600./24.
sxaddpar, header, 'JDMID', jdmid, after='JD'

lst = sxpar(header, 'ST')
lstmid = (ten(lst) + sxpar(header,'EXPTIME')/3600./2.) mod 24
zenithra = 15.*lstmid
zenithdec = 19.82636

  ; these are now using the midpoint of the exposure.  How accurate they are remains to be seen?
moonpos, jdmid, moonra, moondec  ; needs isarray, which seems to have vanished from the GSFC library.
sunpos, jdmid, sunra, sundec    

radeg = 15.*ten(sxpar(header, 'RA'))
decdeg = ten(sxpar(header, 'DEC'))

gcirc, 2, zenithra, zenithdec, moonra, moondec, moonzdist
gcirc, 2, zenithra, zenithdec, sunra, sundec, sunzdist
gcirc, 2, radeg, decdeg, moonra, moondec, moondist

sunelev = 90. - sunzdist/3600.
moonelev = 90. - moonzdist/3600.
moonsep = moondist/3600

sxaddpar, header, 'SUNELEV', sunelev, (sunelev>0.?'Above horizon':'Below horizon'), before='ROTMODE'
sxaddpar, header, 'MOONELEV', moonelev, (moonelev>0?'Above horizon':'Below horizon'), before='ROTMODE'
sxaddpar, header, 'MOONDIST', moonsep, before='ROTMODE'


; --- Dispersion and central wavelength ---

if lmode eq 'spec' then begin
  dispersion = 0.
  if camera eq 'b' then begin
    grism = clip(sxpar(header, 'GRISNAME'))
    if grism eq  '300/5000' then dispersion = 1.43
    if grism eq  '400/3400' then dispersion = 1.09
    if grism eq  '600/4000' then dispersion = 0.63
    if grism eq '1200/3400' then dispersion = 0.24
    if grism eq  '300/5000' then cwav = 4495.
    if grism eq  '400/3400' then cwav = 3600. 
    if grism eq  '600/4000' then cwav = 4305.
    if grism eq '1200/3400' then cwav = 3400.
    if dispersion eq 0. then begin
       print, '   WARNING: Unrecognized grism GRISNAME=', grism 
    endif else begin
       sxaddpar, header, 'DISPERSN', dispersion*ybin, 'Ang/pix'  ; ybin (not rotated yet)
       sxaddpar, header, 'CWAVE',    cwav, 'Expected central wavelength'
    endelse
    sxaddpar, header, 'GRNAME', grism
  endif
  if camera eq 'r' then begin
   grating = clip(sxpar(header,'GRANAME'))
   if grating eq  '150/7500' then dispersion = 3.0
   if grating eq  '300/5000' then dispersion = 1.59
   if grating eq  '400/8500' then dispersion = 1.16
   if grating eq  '600/5000' then dispersion = 0.80
   if grating eq  '600/7500' then dispersion = 0.80
   if grating eq  '600/10500' then dispersion = 0.80
   if grating eq  '831/8200' then dispersion = 0.58
   if grating eq  '900/5500' then dispersion = 0.53
   if grating eq '1200/7500' then dispersion = 0.40
   if grating eq '600/10000' then dispersion = 0.80
   if grating eq '1200/9000' then dispersion = 0.40
   if lrisversion eq 1 then dispersion *= 0.210/0.135
   if dispersion eq 0. then print, '   WARNING: Unrecognized grating GRANAME=', grating $
   else sxaddpar, header, 'DISPERSN', dispersion*ybin, 'Expected dispersion, Ang/pix'  ; ybin (not rotated yet)
   sxaddpar, header, 'GRNAME', grating
   sxaddpar, header, 'CWAVE',  sxpar(header,'WAVELEN'), 'Expected central wavelength'  ; + standard offset...?
  endif
endif

; --- Astrometric WCS keywords ---

ra = ten(strtrim(sxpar(header,'RA'),2))*15  ; hours to degrees
dec = ten(strtrim(sxpar(header,'DEC'),2))
padeg = sxpar(header, 'ROTPOSN')
parad = padeg*3.1415927/180.
pxscaledeg = pxscale/3600.
parity = 1
if lrisversion eq 1 and camera eq 'r' then parity = -1 
if lrisversion eq 1 and camera eq 'r' then parad += !pi/2
; if lrisversion eq 1 and camera eq 'b' then parad += !pi 
if lmode eq 'img' then begin
   sxaddpar, header, 'CD1_1',  pxscaledeg * cos(parad)*parity
   sxaddpar, header, 'CD1_2',  pxscaledeg * sin(parad)
   sxaddpar, header, 'CD2_1', -pxscaledeg * sin(parad)*parity
   sxaddpar, header, 'CD2_2',  pxscaledeg * cos(parad)
   sxaddpar, header, 'CRVAL1', ra
   sxaddpar, header, 'CRVAL2', dec
   sxaddpar, header, 'CTYPE1', 'RA---TAN'
   sxaddpar, header, 'CTYPE2', 'DEC--TAN'
   ; crpix will wait until post-cropping
endif

; --- Spectroscopic WCS keywords ---

if lmode eq 'spec' then begin
   sxaddpar, header, 'CTYPE1', 'WAVE-WAV'
   sxaddpar, header, 'CRPIX1', 2048./ybin  ; might need to reverse xbin/ybin
   sxaddpar, header, 'CRVAL1', sxpar(header, 'CWAVE')
   sxaddpar, header, 'CRUNIT1', 'Angstrom'
   sxaddpar, header, 'CD1_1', sxpar(header, 'DISPERSN')

   sxaddpar, header, 'CRPIX2', 2048/xbin  ; This is the TOP of the CCD (on right chip).
                                          ; Ideally the nominal trace position should be put in here.
   sxaddpar, header, 'CTYPE2', 'DEC---TAN'
   sxaddpar, header, 'CRVAL2', 0.   ; just give the relative offset
   sxaddpar, header, 'CD2_2', pxscaledeg*3600.  ; use arcseconds for convenience
endif



headtime = systime(/seconds)  ; mark time after header modifications, including epheremis



; -------- Fix CCD bug (?) -----------

; There is some bug in the binning procedure itself that produces a blank line 
; next to the chip gap in binned data.  On the red it seems to place two lines
; on the right and on the blue it seems to produce one line each on blue and red.

; Fix it by shifting the image to remove this stripe.

; The "ground truth" standard here is the array produced in 1x1 binning by readhmdu.
; Images taken in 2x2 should be identical but resampled.

; correct for the different pixel readout pattern in binned data (some bug on the CCD?)
; Maybe this should be moved over to readmududap.

if lrisversion eq 3 and camera eq 'r' and xbin eq 2 then begin
   ampl2 = fix(strsplit(sxpar(header,'AMP1'),',',/extract))
   ampr1 = fix(strsplit(sxpar(header,'AMP4'),',',/extract))
   ; hold = array[nxin-2:nxin-1,*]
   array[ampr1[0]-1:nxin-3,*] = array[ampr1[0]+1:nxin-1,*] ; shift right half by 2 pixels
   ; array[nxin-2:nxin-1,*] = hold ; not actually sure if this is the best thing to do -
                                   ; will mess up the edge of an imaging frame windowed and
                                   ; binned for spectroscopy.
                           
endif
if camera eq 'b' and xbin eq 2 then begin
   ampl2 = fix(strsplit(sxpar(header,'AMP2'),',',/extract)) 
   ampr1 = fix(strsplit(sxpar(header,'AMP3'),',',/extract))
   array[ampr1[0]-1:nxin-3,*] = array[ampr1[0]+1:nxin-1,*] ; shift right half by 2 (binned?) pixels

   array = shift(array,1,0)  ; shift EVERYTHING by 1 pixels because there is also an offset across the whole field...
endif



; New exciting bug discovered - diminished columns (5-10% lower response) seem to move around
; even within the same setting.


; -------------- Figure out the amp/chip boundaries ------------


if camera eq 'b' then begin
   ampl1 = fix(strsplit(sxpar(header,'AMP1'),',',/extract))
   ampl2 = fix(strsplit(sxpar(header,'AMP2'),',',/extract))
   ampr1 = fix(strsplit(sxpar(header,'AMP3'),',',/extract))
   ampr2 = fix(strsplit(sxpar(header,'AMP4'),',',/extract))
endif else begin
   if lrisversion ge 2 then begin
     ampl1 = fix(strsplit(sxpar(header,'AMP2'),',',/extract))
     ampl2 = fix(strsplit(sxpar(header,'AMP1'),',',/extract))  ; red amps are not in sequence
     ampr1 = fix(strsplit(sxpar(header,'AMP4'),',',/extract))
     ampr2 = fix(strsplit(sxpar(header,'AMP3'),',',/extract))
   endif else begin
     amp1 = fix(strsplit(sxpar(header,'AMP1'),',',/extract))
     amp2 = fix(strsplit(sxpar(header,'AMP2'),',',/extract))
   endelse
endelse

if camera eq 'b' or lrisversion gt 2 then begin
  chipl = [min([ampl1,ampl2]), max([ampl1,ampl2])]
  chipr = [min(ampr1),         max([ampr1,ampr2])]

  chipboundaryl = chipl[1]
  chipboundaryr = chipr[0] ; should be separated by one...
endif else begin
  chip = [amp1[0] < amp2[0], amp1[1] > amp2[1]]
endelse


; ---------------- Flag/fix bad pixels -------------------


if camera eq 'r' then begin


  ; This block is not actually being used by the pipeline anymore.
  if keyword_set(flag) or keyword_set(fixdefect) then begin
    ; Load bad pixel mask and rebin/crop appropriately
   
    reusemask = 0
    if n_elements(mask) gt 0 and n_elements(prevversion) gt 0 and n_elements(prevpane) eq n_elements(pane) then begin
      if lrisversion eq prevversion and xbin eq prevxbin and ybin eq prevybin and total(pane eq prevpane) eq n_elements(pane) then $
         reusemask = 1      
    endif
    ; reuse the cropped mask if we can, although this saves only trivial processing time (aside from the time saved by not rereading the file)
 
    if reusemask eq 0 then begin ; load mask from disk
       ; should confirm that this works as planned....
       bpmfile = ''
       if camera eq 'r' and lrisversion eq 1 and n_elements(bpm1) lt 2 then bpmfile = bpmpath+'bpm_red1.fits'
       if camera eq 'r' and lrisversion eq 2 and n_elements(bpm2) lt 2 then bpmfile = bpmpath+'bpm_red2.fits'
       if camera eq 'r' and lrisversion eq 3 and n_elements(bpm3) lt 2 then bpmfile = bpmpath+'bpm_red3.fits'
 
       if bpmfile ne '' then if file_test(bpmfile) eq 0 then begin
          print
          print, 'Bad pixel mask '+bpmfile+' not found: check refdata directory'
       endif else begin
          if camera eq 'r' and lrisversion eq 1 and n_elements(bpm1) lt 2 then bpm1 = mrdfits(bpmfile,0,hbpm, /silent)
          if camera eq 'r' and lrisversion eq 2 and n_elements(bpm2) lt 2 then bpm2 = mrdfits(bpmfile,0,hbpm, /silent)
          if camera eq 'r' and lrisversion eq 3 and n_elements(bpm3) lt 2 then bpm3 = mrdfits(bpmfile,0,hbpm, /silent)
          if camera eq 'r' and lrisversion eq 1 then mask = bpm1[pane[0]:pane[0]+pane[2]-1,pane[1]:pane[1]+pane[3]-1]  ; pane is before binning
          if camera eq 'r' and lrisversion eq 2 then mask = bpm2[pane[0]:pane[0]+pane[2]-1,pane[1]:pane[1]+pane[3]-1]
          if camera eq 'r' and lrisversion eq 3 then mask = bpm3[pane[0]:pane[0]+pane[2]-1,pane[1]:pane[1]+pane[3]-1]
          sbpm = size(mask)
          if xbin gt 1 or ybin gt 1 then mask = rebin(mask, sbpm[1]/xbin, sbpm[2]/ybin)
          ;if xbin gt 1 or ybin gt 1 then mask = shift(mask, 1, 0) ; mysterious why this is necessary.  should be revisited with flats.
          ; should no longer be necessary with the problem now identified as a CCD bug, but revisit this to check someday
       endelse
    endif else begin
       ; just reuse the old mask that's already in there.
       sbpm = size(mask)
    endelse
    
    if keyword_set(flag) then array[where(mask gt 0)] = !values.f_nan
    if keyword_set(fixdefect) then begin
      array[where(mask gt 0)] = !values.f_nan
      ct = 1
      while ct gt 0 do begin 
        bad = where(finite(array) eq 0, ct)
        for m = 0, ct-1 do begin
          mm = bad[m]
          array[mm] = median(array[mm-2:mm+2])
        endfor
      endwhile
    endif

  endif

   ; would produce better results to do this after flat-fielding since there is horizontal structure in the response, but this is easier.
   ; Also, not every dropcolumn is affected every time.
   ; May want to consider running this for spectra also.
  if lrisversion eq 3 and lmode eq 'img' and xbin eq 1 then begin   ; does not support binning
      if n_elements(dropcolumns) eq 0 then begin
         readcol, bpmpath+'lrisred3dropcolumns.dat', dropcolumns, format='i', comment='#'
         dropcolumns -= 1  ; convert DS9 to array coordinate
      endif
      for i = 0, n_elements(dropcolumns)-1 do begin
          x = dropcolumns[i]
          for xm = x-1, x-5 do if total(xm eq dropcolumns) eq 0 then break ; find the nearest column that is not also a dropcolumn
          for xp = x+1, x+5 do if total(xp eq dropcolumns) eq 0 then break
          ax = x - pane[0]
          axm = xm - pane[0]
          axp = xp - pane[0]
          if axp gt (size(array)) [1] then continue
          coldiff = median((array[axm,*]+array[axp,*])/2 - array[ax, *],  /even) 
          array[ax,*] += coldiff ;;;;+ !values.f_nan
      endfor
 
      ; also correct the hot pixel cluster while we're at it.
      if (size(array)) [1] gt 1190-pane[0] and (size(array)) [2] gt 1710-pane[1] then $
         array[1185-pane[0]:1190-pane[0],1705-pane[1]:1710-pane[1]] = !values.f_nan   ; corners are probably ok...

      ; also flag some pixels flanking the chip gap.
      if (size(array)) [1] gt 2056-pane[0] then $
         array[2043-pane[0]:2056-pane[0],*] = !values.f_nan
  endif
 
  ; flag the bad columns close to the target position for spectroscopy
  if lrisversion eq 3 and lmode eq 'spec' then begin
     array[(2278-pane[0])/xbin,*] = !values.f_nan
     array[(2276-pane[0])/xbin,*] = !values.f_nan

     ; flag another bad column not close to the trace that messes up the profiles
     array[(2498-pane[0])/xbin,*] = !values.f_nan
  endif

  if lrisversion eq 1 and camera eq 'r' then begin
     s = size(array)
     nx = s[1]
     ny = s[2]
     badtrailx = ([863, 997, 1001]-1-40-pane[0])/xbin
     badtraily = ([723, 1484, 1486]-1-pane[1])/ybin
     for i = 0, n_elements(badtrailx)-1 do $
        if badtrailx[i] ge 0 and badtraily[i] ge 0 and badtrailx[i] lt nx and badtraily[i] lt ny then $
           array[badtrailx[i], 0:badtraily[i]] = !values.f_nan
     badx = ([995, 992]-1-40-pane[0])/xbin
     bady = ([1483, 1477]-1-pane[1])/ybin
     for i = 0, n_elements(badx)-1 do $
        if badx[i] ge 0 and bady[i] ge 0 and badx[i] lt nx and bady[i] lt ny then $
           array[badx[i], bady[i]] = !values.f_nan
  endif


   ; Recognize and replace hyper-saturated pixels (so saturated they register as dead)
   ; from neon or argon lamps
   ;  (could extend this to all files at the expense of processing time if it seems
   ;   necessary)
  if camera eq 'b' or lrisversion gt 1 then begin ; this is in an r camera if block so the camera='b' is impossible
    lamps = sxpar(header, 'LAMPS', count=lampct)
    lamparr = fix(strsplit(lamps,',',/extract))
    if n_elements(lamparr) ge 3 then begin ;  lately this keyword has been missing from some(all?) frames?
      if lamparr[1] or lamparr[2] then begin
        truesaturation = max(array) ; both chips should saturate at the same place?

        arrayl = array[chipl[0]:chipl[1]-4,*] 
        arrayr = array[chipr[0]+4:chipr[1],*] 
        arrayl = fixhypersaturation(arrayl, deadthresh=1500)
        arrayr = fixhypersaturation(arrayr, deadthresh=500)
        array[chipl[0]:chipl[1]-4,*] = arrayl
        array[chipr[0]+4:chipr[1],*] = arrayr

      endif
    endif
  endif

  prevversion = lrisversion
  prevxbin = xbin
  prevybin = ybin
  prevpane = pane
   
endif

prevcamera = camera

flagtime = systime(/seconds)  ; mark time after flagging



; ------------- Gain correction -------------------

if n_elements(gain) eq 0 then gain = [1.,1,1,1]

if camera eq 'r' and lrisversion gt 1 then sxaddpar, header, 'GAIN', 0.95/gain[2], after='RDNOISE' ; is this actually true?  It is close to 1, at any rate.
if camera eq 'r' and lrisversion eq 1 then sxaddpar, header, 'GAIN', 1.98, after='RDNOISE'
if camera eq 'b' then sxaddpar, header, 'GAIN', 1.63/gain[2], after='RDNOISE'
if camera eq 'b' or lrisversion gt 1 then begin
  if max(ampl1) gt 0 then array[ampl1[0]:ampl1[1],*] = array[ampl1[0]:ampl1[1],*] * gain[0]
  if max(ampl2) gt 0 then array[ampl2[0]:ampl2[1],*] = array[ampl2[0]:ampl2[1],*] * gain[1]
  if max(ampr1) gt 0 then array[ampr1[0]:ampr1[1],*] = array[ampr1[0]:ampr1[1],*] * gain[2]
  if max(ampr2) gt 0 then array[ampr2[0]:ampr2[1],*] = array[ampr2[0]:ampr2[1],*] * gain[3]
  if max(ampl1) gt 0 then sxaddpar, header, 'INGAINL1', gain[0]
  if max(ampl2) gt 0 then sxaddpar, header, 'INGAINL2', gain[1]
  if max(ampr1) gt 0 then sxaddpar, header, 'INGAINR1', gain[2]
  if max(ampr2) gt 0 then sxaddpar, header, 'INGAINR2', gain[3]
endif else begin
  if max(amp1) gt 0 and amp1[1] gt amp1[0] then array[amp1[0]:amp1[1],*] = array[amp1[0]:amp1[1],*] * gain[0]
  if max(amp2) gt 0 and amp2[1] gt amp2[0] then array[amp2[0]:amp2[1],*] = array[amp2[0]:amp2[1],*] * gain[1]
  if max(amp1) gt 0 and amp1[1] gt amp1[0] then sxaddpar, header, 'INGAIN1', gain[0]
  if max(amp2) gt 0 and amp2[1] gt amp2[0] then sxaddpar, header, 'INGAIN2', gain[1]
endelse

; ---------- Figure out crop boundaries ----------

if n_elements(ycrop) eq 0 then ycrop = [0,nyin-1]

if n_elements(crops) eq 1 then begin
  if crops eq 'none' then begin
     if camera eq 'blue' or lrisversion gt 1 then begin
       lcrop = chipl
       rcrop = chipr
       ycrop = [0, nyin-1]
       xcrop = [chipl[0], chipr[1]]
     endif else begin
       xcrop = chip
       ycrop = [0, nyin-1]
     endelse
  endif
  if crops eq 'auto' then begin
     ycrop = [0, nyin-1] ; default
     if camera eq 'b' then begin
        if lmode eq 'img' then begin
           lcrop = [(360-pane[0])/xbin > chipl[0], chipl[1]]
           rcrop = [chipr[0], (3790-pane[0])/xbin-1 < chipr[1]]
           ycrop = [(760-pane[1])/ybin > 0, (3200-pane[1])/ybin-1 < nyin-1]
           xcrop = [lcrop[0], rcrop[1]]
        endif
        if lmode eq 'spec' then begin
           lcrop = [(1440-pane[0])/xbin > chipl[0],      chipl[1]]
           rcrop = [chipr[0],      (2670-pane[0])/xbin-1 < chipr[1]]
           xcrop = [lcrop[0], rcrop[1]]
        endif
     endif
     if camera eq 'r' then begin
        if lrisversion eq 1 and lmode eq 'img' then begin
           xcrop = [(116-pane[0])/xbin > chip[0], (1745-pane[0])/xbin-1 < chip[1]]
           ycrop = [0, nyin-1]
        endif
        if lrisversion eq 1 and lmode eq 'spec' then begin
           xcrop = [chip[0], chip[1]] ; no cropping
           ycrop = [(580-pane[1])/ybin > 0, (1500-pane[1])/ybin-1 < nyin-1]
        endif
        if (lrisversion eq 2 or lrisversion eq 3) and lmode eq 'img' then begin
           lcrop = [(430-pane[0])/xbin > chipl[0], chipl[1]]
           rcrop = [chipr[0],  (3778-pane[0])/xbin-1 < chipr[1]]
           ycrop = [(638-pane[1])/ybin > 0, (3100-pane[1])/ybin-1 < nyin-1]
           xcrop = [lcrop[0], rcrop[1]]
        endif
        if (lrisversion eq 2 or lrisversion eq 3) and lmode eq 'spec' then begin
           lcrop = [(1520-pane[0])/xbin > chipl[0], chipl[1]]
           rcrop = [chipr[0],  (2660-pane[0])/xbin-1 < chipr[1]] ; 2660
           xcrop = [lcrop[0], rcrop[1]]
        endif
     endif
  endif
endif

dotpos = strpos(filename,'.',/reverse_search)
slashpos = strpos(filename,'/',/reverse_search)
fileroot = strmid(filename,slashpos+1,dotpos-slashpos-1)
fileext = strmid(filename,dotpos+1)

if merge or multi then begin
   if n_elements(outname) gt 0 then outfilename = outname
   if n_elements(outfilename) eq 0 then begin
      if n_elements(prefix) eq 0 and n_elements(suffix) eq 0 then begin
         prefix = 'p'
      endif
      if n_elements(prefix) gt 0 then outfilename = prefix+fileroot+'.'+fileext
      if n_elements(suffix) gt 0 then outfilename = fileroot+suffix+'.fits'
   endif
endif


get_date, now
hist = 'Processed by lrisprepare '+now

if ycrop[1] le ycrop[0] or xcrop[1] le xcrop[0] then begin
   print
   print, '  ERROR:  Invalid crop region.  Cannot prepare file.'
   ; stop
   return
   ; This seems to occur when processing MIRA files.
   ; investigate why, someday.  2012-04-27 run.
endif

; deal with crate crashes - doesn't seem to work though
if max(array) le 0 then begin
   print, '  File contains no positive data!  Not preparing.'
   return
endif

; ------ Do the cropping and make final edits to the header, then write to disk ------

sxaddpar, header, 'LTM1_1', 1./xbin
sxaddpar, header, 'LTM2_2', 1./ybin



if lrisversion eq 1 and camera eq 'r' then begin
   ; write the file out now for LRIS-R1 (which has only one chip) and stop.

   array  = array[xcrop[0]:xcrop[1],  ycrop[0]:ycrop[1]]
   sxaddpar, header, 'LTV1', -pane[0]*1.0/xbin-xcrop[0] ;
   sxaddpar, header, 'LTV2', -pane[1]*1.0/ybin-ycrop[0] ;
     ; it's not really possible to align the two in DS9 because DS9 does not support fractional LTV offsets and the offsets are in
     ; binned pixels.  So with the different array origins and center vs. edge pixel definitions fractional pixel offsets can't be
     ; avoided.  This might also be going on with the other cameras?

   if n_elements(outname) gt 0 then outfilename = outname
   if n_elements(outfilename) eq 0 then begin
      if n_elements(prefix) eq 0 and n_elements(suffix) eq 0 then begin
         prefix = 'p'
      endif
      if n_elements(prefix) gt 0 then outfilename = prefix+fileroot+'.'+fileext
      if n_elements(suffix) gt 0 then outfilename = fileroot+suffix+'.fits'
   endif
   sxaddpar, header, 'CRPIX1', 1024/xbin - (pane[0]*1.0/xbin+xcrop[0])
   sxaddpar, header, 'CRPIX2', 1024/ybin - (pane[1]*1.0/ybin+ycrop[0])
   if lmode eq 'img' and clip(sxpar(header,'PONAME')) eq 'LRISB' then begin
     vshift = (30/0.135)
     xyad, header, sxpar(header,'CRPIX1'), sxpar(header,'CRPIX2')+vshift, va, vd
     sxaddpar, header, 'CRVAL1', va
     sxaddpar, header, 'CRVAL2', vd
   endif

   if camera eq 'b' and lmode eq 'spec' then countsval = percentile(array,80) $   ; could speed this up a bit...
                                        else countsval = median(array)
   sxaddpar, header, 'COUNTS', countsval
   sxaddpar, header, 'MAXCOUNT', max(array) ; maximum value in the image

   mwrfits, array, outfilename, header, /create
   print, ' -> ', outfilename
   return
endif


if merge then begin
   array  = array[xcrop[0]:xcrop[1],  ycrop[0]:ycrop[1]]
   croptime = systime(/seconds)  ; mark time after cropping

   sxaddpar, header, 'LTV1', -pane[0]*1.0/xbin-xcrop[0] ; LTV1 and LTV2 will be reversed for spectroscopy
   sxaddpar, header, 'LTV2', -pane[1]*1.0/ybin-ycrop[0]

   ; For consistency, we will always define CRPIX to be the center of the near-right amplifier.
   if lmode eq 'img' then begin
     if camera eq 'r' then begin
        sxaddpar, header, 'CRPIX1', 2560/xbin - (pane[0]*1.0/xbin+xcrop[0]) , after='CRVAL2'
        sxaddpar, header, 'CRPIX2', 2048/ybin - (pane[1]*1.0/ybin+ycrop[0]) , after='CRPIX1'
     endif
     if camera eq 'b' then begin
        sxaddpar, header, 'CRPIX1', 2560/xbin - (pane[0]*1.0/xbin+xcrop[0]), after='CRVAL2'
        sxaddpar, header, 'CRPIX2', 2048/ybin - (pane[1]*1.0/ybin+ycrop[0]), after='CRPIX1'
     endif
     vshift = (30/0.135) ; the right CCD center is shifted relative to the blue (and off-center in general)
     xyad, header, sxpar(header,'CRPIX1'), sxpar(header,'CRPIX2')+vshift, va, vd
     sxaddpar, header, 'CRVAL1', va
     sxaddpar, header, 'CRVAL2', vd
   endif

   ampl1 -= xcrop[0]
   ampl2 -= xcrop[0]
   ampr1 -= xcrop[0]
   ampr2 -= xcrop[0]
   if n_elements(ampl1) ge 2 then sxaddpar, header, 'AMPL1', clip(ampl1[0])+','+clip(ampl1[1]), ' (image coordinates)'
   if n_elements(ampl2) ge 2 then sxaddpar, header, 'AMPL2', clip(ampl2[0])+','+clip(ampl2[1])
   if n_elements(ampr1) ge 2 then sxaddpar, header, 'AMPR1', clip(ampr1[0])+','+clip(ampr1[1])
   if n_elements(ampr2) ge 2 then sxaddpar, header, 'AMPR2', clip(ampr2[0])+','+clip(ampr2[1])
   sxdelpar, header, 'AMP1'
   sxdelpar, header, 'AMP2'
   sxdelpar, header, 'AMP3'
   sxdelpar, header, 'AMP4'

   ; sxaddpar, header, 'CRPIX2', ampr1[0]  ; set the WCS y=0 to the chip gap - no?? this makes no sense

   if camera eq 'b' and lmode eq 'spec' then countsval = percentile(array,80) $   ; could speed this up a bit...
                                        else countsval = median(array)
   sxaddpar, header, 'COUNTS', countsval
   sxaddpar, header, 'MAXCOUNT', max(array) ; maximum value in the image

   if camera eq 'r' then sxaddpar, header, 'RDNOISE', 4

   if keyword_set(notranspose) eq 0 and lmode eq 'spec' then begin
      ltranspose, array, header
   endif

   print, ' -> ', outfilename

   sxaddhist, hist, header

   mwrfits, array, outfilename, header, /create

endif
if split then begin
   rheader = header
   lheader = header

   if keyword_set(leftonly)  and n_elements(lcrop) eq 0 and n_elements(xcrop) eq 2 then lcrop = xcrop
   if keyword_set(rightonly) and n_elements(rcrop) eq 0 and n_elements(xcrop) eq 2 then rcrop = xcrop

   doleft  = keyword_set(rightonly) eq 0
   doright = keyword_set(leftonly)  eq 0
 
   if doleft  then arrayl = array[lcrop[0]:lcrop[1],ycrop[0]:ycrop[1]] 
   if doright then arrayr = array[rcrop[0]:rcrop[1],ycrop[0]:ycrop[1]] 

   croptime = systime(/seconds)  ; mark time after cropping


   if lmode eq 'img' then begin
     if camera eq 'r' then begin
        sxaddpar, rheader, 'CRPIX1', 2560/xbin - (pane[0]*1.0/xbin+rcrop[0]) , after='CRVAL2'
        sxaddpar, rheader, 'CRPIX2', 2048/ybin - (pane[1]*1.0/ybin+ycrop[0]) , after='CRPIX1'
 
        sxaddpar, lheader, 'CRPIX1', 1536/xbin - (pane[0]*1.0/xbin+lcrop[0]) , after='CRVAL2'
        sxaddpar, lheader, 'CRPIX2', 2048/ybin - (pane[1]*1.0/ybin+ycrop[0]) , after='CRPIX1'

        vshift = (30/0.135)
        xyad, rheader, sxpar(rheader,'CRPIX1'), sxpar(rheader,'CRPIX2')+vshift, va, vd
        sxaddpar, rheader, 'CRVAL1', va
        sxaddpar, rheader, 'CRVAL2', vd
        xyad, lheader, sxpar(lheader,'CRPIX1'), sxpar(lheader,'CRPIX2')+vshift, va, vd
        sxaddpar, lheader, 'CRVAL1', va
        sxaddpar, lheader, 'CRVAL2', vd

        leftshift = (-1024/xbin-30./0.135)
        xyad, rheader, sxpar(rheader,'CRPIX1')+leftshift, sxpar(rheader,'CRPIX2'), la, ld
        sxaddpar, lheader, 'CRVAL1', la
        sxaddpar, lheader, 'CRVAL2', ld

     endif
     if camera eq 'b' then begin
        sxaddpar, rheader, 'CRPIX1', 2560/xbin - (pane[0]*1.0/xbin+rcrop[0]), after='CRVAL2'
        sxaddpar, rheader, 'CRPIX2', 2048/ybin - (pane[1]*1.0/ybin+ycrop[0]), after='CRPIX1'

        sxaddpar, lheader, 'CRPIX1', 1536/xbin - (pane[0]*1.0/xbin+lcrop[0]), after='CRVAL2'
        sxaddpar, lheader, 'CRPIX2', 2048/ybin - (pane[1]*1.0/ybin+ycrop[0]), after='CRPIX1'

        leftshift = (-1024/xbin-14./0.135)
        xyad, rheader, sxpar(rheader,'CRPIX1')+leftshift, sxpar(rheader,'CRPIX2'), la, ld
        sxaddpar, lheader, 'CRVAL1', la
        sxaddpar, lheader, 'CRVAL2', ld
        ;leftrot = 0.5*!pi/180.
        ;sxaddpar, lheader, 'CD1_1',  pxscaledeg * cos(parad+leftrot)*parity
        ;sxaddpar, lheader, 'CD1_2',  pxscaledeg * sin(parad+leftrot)
        ;sxaddpar, lheader, 'CD2_1', -pxscaledeg * sin(parad+leftrot)*parity
        ;sxaddpar, lheader, 'CD2_2',  pxscaledeg * cos(parad+leftrot)
     endif
   endif
   if lmode eq 'spec' then begin
      sxaddpar, header, 'CRPIX2', 0.  ; for now, set the WCS y=0 to y=0.
   endif

   if doleft  then sxaddpar, lheader, 'LTV1', -pane[0]/xbin-lcrop[0]  ; LTV1 and LTV2 will be reversed in transpose function
   if doright then sxaddpar, rheader, 'LTV1', -pane[0]/xbin-rcrop[0]

   if doleft  then sxaddpar, lheader, 'LTV2', -pane[1]/ybin-ycrop[0]
   if doright then sxaddpar, rheader, 'LTV2', -pane[1]/ybin-ycrop[0]

   if camera eq 'r' then begin
      if doleft  then sxaddpar, lheader, 'RDNOISE', 4
      if doright then sxaddpar, rheader, 'RDNOISE', 3.79 ; confusing, since the fits header says this is 1-13.
   endif

   if doleft  then sxaddpar, lheader, 'CHIP', 'left'
   if doright then sxaddpar, rheader, 'CHIP', 'right'

   if doleft  then sxaddpar, lheader, 'COUNTS', median(arrayl)
   if doright then sxaddpar, rheader, 'COUNTS', median(arrayr)

   if multi and n_elements(outname) gt 0 then outfilename = outname
   if keyword_set(leftonly) and n_elements(loutfilename) eq 0 and n_elements(outfilename) gt 0 then loutfilename=outfilename
   if keyword_set(rightonly) and n_elements(routfilename) eq 0 and n_elements(outfilename) gt 0 then routfilename=outfilename

   if n_elements(prefix) gt 0 and n_elements(rprefix) eq 0 and keyword_set(rightonly) then rprefix = prefix
   if n_elements(prefix) gt 0 and n_elements(lprefix) eq 0 and keyword_set(rightonly) then lprefix = prefix

   if n_elements(loutname) gt 0 then loutfilename = loutname
   if n_elements(loutfilename) eq 0 then begin
      if n_elements(lprefix) eq 0 and n_elements(lsuffix) eq 0 then begin
        lprefix = 'p'
        lsuffix = 'l'
      endif
      if n_elements(lprefix) gt 0 then loutfilename = lprefix+fileroot+'.'+fileext
      if n_elements(lsuffix) gt 0 then loutfilename = fileroot+lsuffix+'.fits'
   endif
   if n_elements(routname) gt 0 then routfilename = routname
   if n_elements(routfilename) eq 0 then begin
      if n_elements(rprefix) eq 0 and n_elements(rsuffix) eq 0 then begin
         rprefix = 'p'
         rsuffix = 'r'
      endif
      if n_elements(rprefix) gt 0 then routfilename = rprefix+fileroot+'.'+fileext
      if n_elements(rsuffix) gt 0 then routfilename = fileroot+rsuffix+'.fits'
   endif

   if  keyword_set(notranspose) eq 0 and lmode eq 'spec' then  begin
      if doleft  then ltranspose, arrayl, lheader 
      if doright then ltranspose, arrayr, rheader 
   endif

   if multi then begin
      print, ' -> ', outfilename 
   endif else begin $
      if doleft and doright        then print, ' -> ', loutfilename, ',', routfilename
      if doleft and (doright eq 0) then print, ' -> ', loutfilename
      if doright and (doleft eq 0) then print, ' -> ', routfilename
   endelse

   ;if doleft  and file_test(loutfilename) then print, 'Overwriting ', loutfilename
   ;if doright and file_test(routfilename) then print, 'Overwriting ', routfilename

   sxaddhist, hist, lheader
   sxaddhist, hist, rheader

   if multi eq 0 then begin
     if doleft  then mwrfits, arrayl, loutfilename, lheader, /create
     if doright then mwrfits, arrayr, routfilename, rheader, /create
   endif else begin
     if doright then mwrfits, arrayr, outfilename, rheader, /create
     if doleft  then mwrfits, arrayl, outfilename, lheader     
   endelse

endif

finaltime = systime(/seconds)  ; mark time at end of file processing

if keyword_set(timer) then begin
print, 'Total duration:  ', finaltime-starttime
print, '  Read duration: ', readtime-starttime
print, '  Head duration: ', headtime-readtime
print, '  Flag duration: ', flagtime-headtime
print, '  Crop duration: ', croptime-flagtime
print, '  Write duration:', finaltime-croptime
endif

end

